import random

import numpy as np
from lightcurvelynx.base_models import FunctionNode
from lightcurvelynx.models.basic_models import (
    ConstantSEDModel,
    LinearWavelengthModel,
    SinWaveModel,
    StepModel,
)
from lightcurvelynx.utils.wave_extrapolate import ConstantExtrapolation, LinearDecay


def _sampler_fun(magnitude, offset=0.0, **kwargs):
    """Return a random value between 0 and magnitude.

    Parameters
    ----------
    magnitude : float
        The range of brightness magnitude
    offset : float, optional
        The offset.
    **kwargs : dict, optional
        Absorbs additional parameters
    """
    return magnitude * random.random() + offset


def test_constant_sed_model() -> None:
    """Test that we can sample and create a ConstantSEDModel object."""
    model = ConstantSEDModel(brightness=10.0, node_label="my_constant_sed_model")
    state = model.sample_parameters()
    assert model.get_param(state, "brightness") == 10.0
    assert model.get_param(state, "ra") is None
    assert model.get_param(state, "dec") is None
    assert model.get_param(state, "distance") is None
    assert str(model) == "my_constant_sed_model"

    times = np.array([1, 2, 3, 4, 5, 10])
    wavelengths = np.array([100.0, 200.0, 300.0])

    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (6, 3)
    assert np.all(values == 10.0)

    # We can set a value we have already added.
    model.set_parameter("brightness", 5.0)
    values = model.evaluate_sed(times, wavelengths)
    assert values.shape == (6, 3)
    assert np.all(values == 5.0)


def test_constant_sed_model_pytree():
    """Test that the PyTree only contains brightness."""
    model = ConstantSEDModel(brightness=10.0, node_label="my_constant_sed_model")
    state = model.sample_parameters()

    pytree = model.build_pytree(state)
    assert pytree["my_constant_sed_model"]["brightness"] == 10.0
    assert pytree["my_constant_sed_model"]["t0"] is None
    assert len(pytree["my_constant_sed_model"]) == 2
    assert len(pytree) == 1


def test_constant_sed_model_host() -> None:
    """Test that we can sample and create a ConstantSEDModel object with properties
    derived from the host object."""
    host = ConstantSEDModel(brightness=15.0, ra=1.0, dec=2.0, distance=3.0)
    model = ConstantSEDModel(brightness=10.0, ra=host.ra, dec=host.dec, distance=host.distance)
    state = model.sample_parameters()

    assert model.get_param(state, "brightness") == 10.0
    assert model.get_param(state, "ra") == 1.0
    assert model.get_param(state, "dec") == 2.0
    assert model.get_param(state, "distance") == 3.0
    assert str(model) == "ConstantSEDModel_0"

    # Test that we have given a different name to the host.
    assert str(host) == "ConstantSEDModel_1"


def test_constant_sed_model_resample() -> None:
    """Check that we can call resample on the model parameters."""
    model = ConstantSEDModel(brightness=_sampler_fun, magnitude=100.0)

    num_samples = 100
    values = np.zeros((num_samples, 1))
    for i in range(num_samples):
        state = model.sample_parameters()
        values[i] = model.get_param(state, "brightness")

    # Check that the values fall within the expected bounds.
    assert np.all(values >= 0.0)
    assert np.all(values <= 100.0)

    # Check that the values are not all the same.
    assert not np.all(values == values[0])


def test_step_model() -> None:
    """Test that we can sample and create a StepModel object."""
    host = ConstantSEDModel(brightness=150.0, ra=1.0, dec=2.0, distance=3.0)
    model = StepModel(brightness=15.0, t0=1.0, t1=2.0, ra=host.ra, dec=host.dec, distance=host.distance)
    state = model.sample_parameters()

    param_values = model.get_local_params(state)
    assert param_values["brightness"] == 15.0
    assert param_values["t0"] == 1.0
    assert param_values["t1"] == 2.0
    assert param_values["ra"] == 1.0
    assert param_values["dec"] == 2.0
    assert param_values["distance"] == 3.0

    times = np.array([0.0, 1.0, 2.0, 3.0])
    wavelengths = np.array([100.0, 200.0])
    expected = np.array([[0.0, 0.0], [15.0, 15.0], [15.0, 15.0], [0.0, 0.0]])

    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (4, 2)
    assert np.array_equal(values, expected)


def test_step_model_resample() -> None:
    """Check that we can call resample on the model parameters."""
    random.seed(1111)

    model = StepModel(
        brightness=FunctionNode(_sampler_fun, magnitude=100.0),
        t0=0.0,
        t1=FunctionNode(_sampler_fun, magnitude=5.0, offset=1.0),
    )

    num_samples = 1000
    brightness_vals = np.zeros((num_samples, 1))
    t_end_vals = np.zeros((num_samples, 1))
    t_start_vals = np.zeros((num_samples, 1))
    for i in range(num_samples):
        state = model.sample_parameters()
        brightness_vals[i] = model.get_param(state, "brightness")
        t_end_vals[i] = model.get_param(state, "t1")
        t_start_vals[i] = model.get_param(state, "t0")

    # Check that the values fall within the expected bounds.
    assert np.all(brightness_vals >= 0.0)
    assert np.all(brightness_vals <= 100.0)
    assert np.all(t_start_vals == 0.0)
    assert np.all(t_end_vals >= 1.0)
    assert np.all(t_end_vals <= 6.0)

    # Check that the expected values are close.
    assert abs(np.mean(brightness_vals) - 50.0) < 5.0
    assert abs(np.mean(t_end_vals) - 3.5) < 0.5

    # Check that the brightness or end values are not all the same.
    assert not np.all(brightness_vals == brightness_vals[0])
    assert not np.all(t_end_vals == t_end_vals[0])


def test_sin_wave_model() -> None:
    """Test that we can sample and create a SinWaveModel object."""
    model = SinWaveModel(amplitude=15.0, frequency=1.0, t0=0.0)
    state = model.sample_parameters()

    times = np.array([0.0, 1.0 / 12.0, 1.0 / 4.0, 5.0 / 12.0, 0.5])
    wavelengths = np.array([100.0, 200.0])
    expected = np.array([[0.0, 0.0], [7.5, 7.5], [15.0, 15.0], [7.5, 7.5], [0.0, 0.0]])

    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (5, 2)
    assert np.allclose(values, expected)

    model2 = SinWaveModel(brightness=10.0, amplitude=15.0, frequency=1.0, t0=0.0)
    state2 = model2.sample_parameters()
    expected2 = np.array([[10.0, 10.0], [17.5, 17.5], [25.0, 25.0], [17.5, 17.5], [10.0, 10.0]])

    values2 = model2.evaluate_sed(times, wavelengths, state2)
    assert values2.shape == (5, 2)
    assert np.allclose(values2, expected2)


def test_linear_wavelength_model() -> None:
    """Test that we can sample and create a LinearWavelengthModel object."""
    model = LinearWavelengthModel(linear_base=1.0, linear_scale=0.1)
    state = model.sample_parameters()

    times = np.arange(0.0, 10.0, 0.5)
    wavelengths = np.array([0.0, 1000.0, 2000.0])
    expected = np.tile(np.array([1.0, 101.0, 201.0]), (len(times), 1))

    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (20, 3)
    assert np.allclose(values, expected)


def test_linear_wavelength_model_redshift() -> None:
    """Test that we correctly apply a redshift to the wavelengths."""
    model = LinearWavelengthModel(linear_base=1.0, linear_scale=0.1, redshift=0.2, t0=0.0)
    state = model.sample_parameters()

    times = np.arange(0.0, 10.0, 0.5)
    wavelengths = np.array([0.0, 1000.0, 1111.1, 1500.0, 2000.0])
    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (20, 5)

    # The shifted (rest frame) wavelengths are given by the equation:
    #   rest_frame_wavelengths = observer_frame_wavelengths / (1 + redshift)
    # so we evaluate the linear function at those shifted wavelengths. Then scale
    # those fluxes by (1 + redshift) to get back to the observer frame.
    rest_frame_wavelengths = wavelengths / (1.2)
    expected_single_time = 1.2 * (1.0 + 0.1 * rest_frame_wavelengths)
    expected = np.tile(expected_single_time, (len(times), 1))
    assert np.allclose(values, expected)


def test_linear_wavelength_model_bounds() -> None:
    """Test that we correctly apply a redshift to the wavelengths."""
    model = LinearWavelengthModel(linear_base=1.0, linear_scale=0.1, min_wave=1000.0, max_wave=2000.0)
    state = model.sample_parameters()

    times = np.arange(0.0, 10.0, 0.5)
    wavelengths = np.array([500.0, 1000.0, 1500.0, 2000.0, 2500.0])
    values = model.evaluate_sed(times, wavelengths, state)

    # Without any extrapolation, we zero pad the data.
    expected = np.tile(np.array([0.0, 101.0, 151.0, 201.0, 0.0]), (len(times), 1))
    assert np.allclose(values, expected)

    # We fill in with a constant value.
    model2 = LinearWavelengthModel(
        linear_base=1.0,
        linear_scale=0.1,
        min_wave=1000.0,
        max_wave=2000.0,
        wave_extrapolation=ConstantExtrapolation(value=100.0),
    )
    state2 = model2.sample_parameters()
    values2 = model2.evaluate_sed(times, wavelengths, state2)
    expected2 = np.tile(np.array([100.0, 101.0, 151.0, 201.0, 100.0]), (len(times), 1))
    assert np.allclose(values2, expected2)

    # We linearly decay the values.
    model3 = LinearWavelengthModel(
        linear_base=1.0,
        linear_scale=0.1,
        min_wave=1000.0,
        max_wave=2000.0,
        wave_extrapolation=LinearDecay(decay_width=100.0),
    )
    wavelengths3 = np.array([500.0, 950.0, 1000.0, 1500.0, 2000.0, 2050.0, 2300.0])
    state3 = model3.sample_parameters()
    values3 = model3.evaluate_sed(times, wavelengths3, state3)
    expected3 = np.tile(np.array([0.0, 50.5, 101.0, 151.0, 201.0, 100.5, 0.0]), (len(times), 1))
    assert np.allclose(values3, expected3)



import numpy as np
import pytest
from lightcurvelynx.models.static_sed_model import StaticBandfluxModel, StaticSEDModel
from lightcurvelynx.utils.io_utils import write_numpy_data


def test_single_static_sed() -> None:
    """Test that we can create and sample a StaticSEDModel object with a single SED."""
    sed = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    model = StaticSEDModel(sed, node_label="test")
    assert len(model) == 1

    times = np.array([1, 2, 3, 10, 20])
    wavelengths = np.array([50.0, 100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 450.0])

    values = model.evaluate_sed(times, wavelengths)
    assert values.shape == (5, 9)

    expected = np.array([0.0, 10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0, 0.0])
    for t_idx in range(5):
        assert np.array_equal(values[t_idx, :], expected)


def test_static_sed_fail() -> None:
    """Test that we correctly fail on bad SEDs."""
    # Non-numpy arrays
    with pytest.raises(ValueError):
        _ = StaticSEDModel([None], node_label="test")
    with pytest.raises(ValueError):
        _ = StaticSEDModel([1.0], node_label="test")

    # Incorrectly shaped data.
    sed = np.array(
        [
            [100.0, 200.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0],  # fluxes
            [0.0, 0.0, 0.0],  # Other row
        ]
    )
    with pytest.raises(ValueError):
        _ = StaticSEDModel([sed], node_label="test")

    # Wavelengths unsorted.
    sed = np.array(
        [
            [100.0, 200.0, 400.0, 300.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    with pytest.raises(ValueError):
        _ = StaticSEDModel([sed], node_label="test")


def test_static_sed_from_file(tmp_path) -> None:
    """Test that we can create a StaticSEDModel object from a file."""
    test_sed = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    times = np.array([1, 2, 3, 10, 20])
    wavelengths = np.array([50.0, 100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 450.0])
    expected = np.array([0.0, 10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0, 0.0])

    for fmt in ["npy", "npz", "txt", "csv"]:
        file_path = tmp_path / f"test_sed.{fmt}"
        write_numpy_data(file_path, test_sed.T)

        model = StaticSEDModel.from_file(file_path, node_label="test")
        assert len(model) == 1

        values = model.evaluate_sed(times, wavelengths)
        assert values.shape == (5, 9)

        for t_idx in range(5):
            assert np.array_equal(values[t_idx, :], expected)

    # Try an invalid array shape.
    test_sed_invalid = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
            [0.0, 0.0, 0.0, 0.0],  # Other row
        ]
    )
    file_path_invalid = tmp_path / "test_sed.csv"
    write_numpy_data(file_path_invalid, test_sed_invalid.T)

    with pytest.raises(ValueError):
        _ = StaticSEDModel.from_file(file_path_invalid, node_label="test")


def test_multiple_static_seds() -> None:
    """Test that we can create and sample a StaticSEDModel object with a multiple SEDs."""
    sed0 = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    sed1 = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [20.0, 40.0, 40.0, 20.0],  # fluxes
        ]
    )
    model = StaticSEDModel([sed0, sed1], weights=[0.25, 0.75], node_label="test")
    assert len(model) == 2

    # Check that all of the indices are 0 or 1 and the split is approximately 25/75
    params = model.sample_parameters(num_samples=10_000)
    inds_0 = params["test"]["selected_idx"] == 0
    inds_1 = params["test"]["selected_idx"] == 1
    assert np.all(inds_0 | inds_1)
    assert 1_500 < np.count_nonzero(inds_0) < 3_500
    assert 6_500 < np.count_nonzero(inds_1) < 8_500

    times = np.array([1, 2, 3])
    wavelengths = np.array([50.0, 100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 450.0])
    values = model.evaluate_sed(times, wavelengths, params)

    expected_0 = np.tile([0.0, 10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0, 0.0], 3).reshape(3, 9)
    expected_1 = 2.0 * expected_0

    for idx in range(10_000):
        if inds_0[idx]:
            assert np.array_equal(values[idx], expected_0)
        else:
            assert np.array_equal(values[idx], expected_1)


def test_multiple_static_seds_min_max():
    """Test that we can get the min and max wavelengths from a StaticSEDModel object with multiple SEDs."""
    sed0 = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    sed1 = np.array(
        [
            [200.0, 300.0, 400.0, 500.0],  # Wavelengths
            [20.0, 40.0, 40.0, 20.0],  # fluxes
        ]
    )
    model = StaticSEDModel([sed0, sed1], weights=[0.5, 0.5], node_label="test")

    states = model.sample_parameters(num_samples=1)

    # Force the selected_idx to be 0 for the test.
    states.set("test", "selected_idx", 0)
    assert model.minwave(states) == 100.0
    assert model.maxwave(states) == 400.0

    # Force the selected_idx to be 1 for the test.
    states.set("test", "selected_idx", 1)
    assert model.minwave(states) == 200.0
    assert model.maxwave(states) == 500.0

    # We fail if we do not pass in the states.
    with pytest.raises(ValueError):
        _ = model.minwave()
    with pytest.raises(ValueError):
        _ = model.maxwave()


def test_single_static_bandflux() -> None:
    """Test that we can create and sample a StaticBandfluxModel object with a single bandflux."""
    bandflux = {
        "r": 10.0,
        "g": 20.0,
        "b": 30.0,
    }
    model = StaticBandfluxModel(bandflux, node_label="test")
    assert len(model) == 1

    times = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    filters = ["r", "r", "g", "b", "r", "g", "b", "g", "b", "r"]
    expected = np.array([10.0, 10.0, 20.0, 30.0, 10.0, 20.0, 30.0, 20.0, 30.0, 10.0])

    state = model.sample_parameters(num_samples=1)
    fluxes = model.evaluate_bandfluxes(None, times, filters, state)
    assert len(fluxes) == 10
    assert np.array_equal(fluxes, expected)


def test_multiple_static_bandflux() -> None:
    """Test that we can create and sample a StaticBandfluxModel object with multiple bandfluxes."""
    bandflux0 = {"r": 10.0, "g": 20.0, "b": 30.0}
    bandflux1 = {"r": 15.0, "g": 25.0, "b": 35.0}
    model = StaticBandfluxModel([bandflux0, bandflux1], weights=[0.25, 0.75], node_label="test")
    assert len(model) == 2

    # Check that all of the indices are 0 or 1 and the split is approximately 25/75
    params = model.sample_parameters(num_samples=10_000)
    inds_0 = params["test"]["selected_idx"] == 0
    inds_1 = params["test"]["selected_idx"] == 1
    assert np.all(inds_0 | inds_1)
    assert 1_500 < np.count_nonzero(inds_0) < 3_500
    assert 6_500 < np.count_nonzero(inds_1) < 8_500

    times = np.array([1, 2, 3, 4, 5])
    filters = ["r", "r", "g", "b", "r"]
    expected0 = np.array([10.0, 10.0, 20.0, 30.0, 10.0])
    expected1 = np.array([15.0, 15.0, 25.0, 35.0, 15.0])

    fluxes = model.evaluate_bandfluxes(None, times, filters, params)
    assert fluxes.shape == (10_000, 5)
    for idx in range(10_000):
        if inds_0[idx]:
            assert np.array_equal(fluxes[idx], expected0)
        else:
            assert np.array_equal(fluxes[idx], expected1)


class DummySynphotModel:
    """A fake synphot model used for testing.

    Attributes
    ----------
    waveset : numpy.ndarray
        The wavelengths at which the SED is defined (in angstroms)
    fluxset : numpy.ndarray
        The flux at each given wavelength (in nJy.)
    """

    def __init__(self, waveset, fluxset):
        self.waveset = waveset
        self.fluxset = fluxset
        self.z = 0.0  # Redshift

    def __call__(self, waves, **kwargs):
        """Return the flux for the given wavelengths as interpolated PHOTLAM.

        Parameters
        ----------
        waves : numpy.ndarray
            The wavelengths at which to evaluate the SED (in angstroms).
        **kwargs : dict
            Additional keyword arguments (ignored).

        Returns
        -------
        numpy.ndarray
            The interpolated flux values at the given wavelengths (in PHOTLAM).
        """
        # Return a dummy SED for the given wavelengths
        return np.interp(waves, self.waveset, self.fluxset, left=0.0, right=0.0)


def test_static_sed_from_synphot() -> None:
    """Test that we can create a StaticSEDModel from a synphot model."""
    # Create a dummy model with 4 samples of SEDs [10.0, 20.0, 30.0, 40.0] in nJy.
    # Since synphot uses PHOTLAM, we preconvert and provide in that unit.
    sp_model = DummySynphotModel(
        waveset=np.array([1000.0, 2000.0, 3000.0, 4000.0]),
        fluxset=np.array([1.50919018e-08, 1.50919018e-08, 1.50919018e-08, 1.50919018e-08]),
    )
    model = StaticSEDModel.from_synphot(sp_model)
    assert len(model) == 1

    times = np.array([1, 2, 3, 10, 20])
    wavelengths = np.array([500.0, 1000.0, 1500.0, 2000.0, 3000.0, 5000.0])
    expected = np.array([0.0, 10.0, 15.0, 20.0, 30.0, 0.0])
    fluxes = model.evaluate_sed(times, wavelengths)
    assert fluxes.shape == (len(times), len(wavelengths))
    for i in range(len(times)):
        np.testing.assert_allclose(fluxes[i, :], expected, rtol=1e-5)

    # We fail is the synphot model has a redshift defined.
    sp_model.z = 0.5
    with pytest.raises(ValueError):
        _ = StaticSEDModel.from_synphot(sp_model)



import importlib

import numpy as np
import pytest
from citation_compass import find_in_citations
from lightcurvelynx import _LIGHTCURVELYNX_TEST_DATA_DIR
from lightcurvelynx.astro_utils.dustmap import ConstantHemisphereDustMap, DustmapWrapper, SFDMap
from lightcurvelynx.math_nodes.given_sampler import GivenValueList


def test_constant_hemisphere_dust_map():
    """Test that we can create and directly query a ConstantHemisphereDustMap object."""
    dust_map = ConstantHemisphereDustMap(north_ebv=0.5, south_ebv=0.4)
    assert dust_map.north_ebv == 0.5
    assert dust_map.south_ebv == 0.4

    # Check that the dust map returns the correct values for different locations.
    assert dust_map.compute_ebv(0.0, 45.0) == 0.5
    assert dust_map.compute_ebv(-10.0, 70.0) == 0.5
    assert dust_map.compute_ebv(5.0, 0.001) == 0.5
    assert dust_map.compute_ebv(0.0, -45.0) == 0.4
    assert dust_map.compute_ebv(20.0, -0.001) == 0.4

    ebvs = dust_map.compute_ebv(
        [0.0, 0.0, 20.0, 340.0, 250.0, 180.0],
        [45.0, -45.0, 15.0, -70.0, -90.0, 0.0],
    )
    assert np.array_equal(ebvs, [0.5, 0.4, 0.5, 0.4, 0.4, 0.5])


def test_dust_map_wrapper():
    """Test that we can create and sample a wrapped dust map."""
    dec_vals = np.arange(-90.0, 90.0, 15.0)
    ra_vals = np.arange(0.0, 360.0, 30.0)
    fake_dust_map = ConstantHemisphereDustMap(north_ebv=0.8, south_ebv=0.1)

    dust_node1 = DustmapWrapper(
        dust_map=fake_dust_map,
        ra=GivenValueList(ra_vals),
        dec=GivenValueList(dec_vals),
        node_label="dust",
    )
    samples = dust_node1.sample_parameters(num_samples=len(dec_vals))
    assert np.all(samples["dust"]["ebv"][dec_vals >= 0] == 0.8)
    assert np.all(samples["dust"]["ebv"][dec_vals < 0] == 0.1)

    # Try adding an adjustment function to the dust map.
    def _ebv_adjust(value):
        return value + 0.1

    dust_node2 = DustmapWrapper(
        dust_map=fake_dust_map,
        ebv_func=_ebv_adjust,
        ra=GivenValueList(ra_vals),
        dec=GivenValueList(dec_vals),
        node_label="dust",
    )
    samples = dust_node2.sample_parameters(num_samples=len(dec_vals))
    assert np.all(samples["dust"]["ebv"][dec_vals >= 0] == 0.9)
    assert np.all(samples["dust"]["ebv"][dec_vals < 0] == 0.2)


def test_sfdmap():
    """Test that we can create and sample an SFDObject."""
    if importlib.util.find_spec("sfdmap2") is None:
        pytest.skip("sfdmap2 is not installed, skipping SFDMap tests.")

    # Use data from a fake map with zero dust in the high North latitudes
    # and high dust in the South.
    dust = SFDMap(
        _LIGHTCURVELYNX_TEST_DATA_DIR / "dustmaps" / "sfdmap2",
        ra=GivenValueList([125.0, -45.0, 10.0]),
        dec=GivenValueList([60.0, -30.0, -45.0]),
        node_label="dust",
    )

    # Test that we can call the compute_ebv function directly with specific coordinates
    # (scalar and array) without sampling RA and dec.
    assert dust.compute_ebv(125.0, 60.0) == pytest.approx(0.0, abs=0.001)
    assert dust.compute_ebv(-45.0, -30.0) == pytest.approx(86.0, abs=0.001)

    ebvs = dust.compute_ebv(np.array([125.0, -45.0, -10.0]), np.array([60.0, -30.0, 70.0]))
    assert np.allclose(ebvs, [0.0, 86.0, 0.0], atol=0.001)

    # Test that we can sample SFDMap as a ParameterNode, this will pull
    # RA and Dec values from the given lists.

    samples = dust.sample_parameters(num_samples=3)
    assert np.allclose(samples["dust"]["ebv"], [0.0, 86.0, 86.0], atol=0.01)

    # We fail if we try to create a SFDMap without RA and dec setters.
    with pytest.raises(ValueError):
        _ = SFDMap(_LIGHTCURVELYNX_TEST_DATA_DIR / "dustmaps" / "sfdmap2")


def test_dustmap_citation():
    """Test the citations for the DustMapWrapper model."""
    fake_dust_map = ConstantHemisphereDustMap(north_ebv=0.8, south_ebv=0.1)
    _ = DustmapWrapper(dust_map=fake_dust_map, ra=0.0, dec=0.0)

    citations = find_in_citations("DustmapWrapper")
    for citation in citations:
        assert "Green 2018, JOSS, 3(26), 695" in citation
        assert "https://github.com/gregreen/dustmaps" in citation



import numpy as np
from astropy import constants
from astropy import units as u

# Physics Constants --------------------------------------

# Mass of the sun in grams
M_SUN_G = constants.M_sun.cgs.value
"""Mass of the sun in grams."""

PARSEC_TO_CM = (1 * u.pc).to_value(u.cm)
"""Conversion factor from parsec to cm."""

GAUSS_EFF_AREA2FWHM_SQ = np.pi / (2 * np.log(2))  # ~2.266
"""Effective area of symmetric Gaussian to FWHM squared conversion factor.

It is roughly 2.266, see
https://smtn-002.lsst.io/v/OPSIM-1171/index.html.

Notes
-----
This is derived from two facts for a symmetric 2D Gaussian:

1. FWHM² = 8 * ln(2) * sigma², where sigma is the standard deviation.
2. The convolution of a Gaussian with itself at position of (x,y)=μ̄ is
   1 / sigma² / (4 π).

   E.g.::

       ∫ g²(x,y) dx dy = 1 / sigma² / (4π),

   where::

       g(x,y) = 1 / (2π sigma²) exp(-(x²+y²)/2/sigma²).
"""

ANGSTROM_TO_CM = (1.0 * u.AA).to_value(u.cm)
"""Conversion factor from Angstrom to cm, 1e-8."""

CGS_FNU_UNIT_TO_NJY = (1.0 * u.erg / u.second / u.cm**2 / u.Hz).to_value(u.nJy)
"""Conversion factor from erg/s/cm²/Hz to nJy, 1e32."""


# Plotting constants ----------------------------------------

lsst_filter_plot_colors = {
    "u": "purple",
    "g": "blue",
    "r": "green",
    "i": "yellow",
    "z": "orange",
    "y": "red",
    "LSST_u": "purple",
    "LSST_g": "blue",
    "LSST_r": "green",
    "LSST_i": "yellow",
    "LSST_z": "orange",
    "LSST_y": "red",
    "F062": "blue",
    "F087": "purple",
    "F106": "green",
    "F129": "orange",
    "F146": "gray",
    "F158": "orangered",
    "F184": "magenta",
    "F213": "red",
}
"""The default colors for plotting to match.

We provide values for the filters with and without the "LSST_" prefix.
https://community.lsst.org/t/lsst-filter-profiles/1463
"""



"""An effect that adds microlensing magnification."""

import numpy as np
from citation_compass import CiteClass

from lightcurvelynx.effects.effect_model import EffectModel
from lightcurvelynx.math_nodes.given_sampler import BinarySampler


class Microlensing(EffectModel, CiteClass):
    """A simple microlensing effect that can be applied to basic models. For more
    complex models, such as those blended by the microlensing event, it is recommended
    to create a new physical model.

    This model is a pure Paczynski/point-source point-lens (PSPL) model, without any
    additional effects. They can be added later. At the moment no values are parametrized,
    but they can be in the future.

    Values that should be parametrized are:
    * lens_mass (M_L)
    * lens_distance (D_L)
    * source_distance (D_S)
    * relative_proper_motion (mu_rel, mu_LS)
    * source_radius (R_S)

    References
    ----------
    * V. Bozza, MNRAS 408 (2010) 2188: general algorithm for binary lensing;
    * V. Bozza, E. Bachelet, F. Bartolic, T. Heintz, A. Hoag, M. Hundertmark, MNRAS 479 (2018) 5157:
    BinaryMag2 function, Extended-Source-Point-Lens methods;
    * V. Bozza, E. Khalouei and E. Bachelet, MNRAS 505 (2021) 126: astrometry, generalized limb darkening,
    Keplerian orbital motion;
    * V. Bozza, v. Saggese, G. Covone, P. Rota & J. Zhang, A&A 694 (2025) 219: multiple lenses.

    Attributes
    ----------
    VBM : VBMicrolensing
        The microlensing model.

    Parameters
    ----------
    microlensing_t0 : parameter
        The time of the microlensing peak, in days.
    u_0 : parameter
        Impact parameter, distance from the source at time
        of peak in Einstein radii.
    t_E : parameter
        Einstein timescale, in days.
    probability : float, optional
        The probability of the microlensing event occurring. Default: 1.0
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, microlensing_t0, u_0, t_E, probability=1.0, **kwargs):
        super().__init__(**kwargs, rest_frame=False)
        self.add_effect_parameter("microlensing_t0", microlensing_t0)
        self.add_effect_parameter("u_0", u_0)
        self.add_effect_parameter("t_E", t_E)

        # Add a parameter that indicates whether or not we apply microlensing that
        # is drawn from a distribution with the given probability.
        if probability < 1.0:
            self.add_effect_parameter("apply_microlensing", BinarySampler(probability))
        else:
            # Don't bother to add another node if we will always apply it.
            self.add_effect_parameter("apply_microlensing", True)

        # Create the microlensing model once.
        try:
            import VBMicrolensing
        except ImportError as err:
            raise ImportError(
                "VBMicrolensing package is not installed be default. To use the microlensing effect, "
                "please install it. For example, you can install it with `pip install VBMicrolensing`."
            ) from err
        self.VBM = VBMicrolensing.VBMicrolensing()

    def apply(
        self,
        flux_density,
        times=None,
        wavelengths=None,
        *,
        apply_microlensing=None,
        microlensing_t0=None,
        u_0=None,
        t_E=None,
        rng_info=None,
        **kwargs,
    ):
        """Apply the effect to observations (flux_density values).

        Parameters
        ----------
        flux_density : numpy.ndarray
            A length T X N matrix of flux density values (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD). Not used for this effect.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms). Not used for this effect.
        apply_microlensing : bool
            Whether or not to apply microlensing.
        microlensing_t0 : float
            The time of the microlensing peak, in days.
        u_0 : float
            Impact parameter, distance from the source at time
            of peak in Einstein radii.
        t_E : float
            Einstein timescale, in days.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of flux densities after the effect is applied (in nJy).
        """
        # Skip the microlensing application if needed.
        if not apply_microlensing:
            return flux_density

        if microlensing_t0 is None:
            raise ValueError("microlensing_t0 must be provided")
        if u_0 is None:
            raise ValueError("u_0 must be provided")
        if t_E is None:
            raise ValueError("t_E must be provided")

        # Array of parameters
        # Note that VBMicrolensing requires some parameters
        # in log space, as per example:
        # https://github.com/valboz/VBMicrolensing/blob/main/docs/python/LightCurves.md
        pr = [np.log(u_0), np.log(t_E), microlensing_t0]

        # Calculates the PSPL magnification at different times with parameters in pr
        # Attention: VBMicrolensing assumes that times is an np.array or a list.
        vbm_results = self.VBM.PSPLLightCurve(pr, times)

        # array of magnifications at each time in time_stamp
        magnifications = np.asarray(vbm_results[0])

        flux_density = flux_density * magnifications[:, np.newaxis]
        return flux_density

    def apply_bandflux(
        self,
        bandfluxes,
        *,
        apply_microlensing=None,
        microlensing_t0=None,
        u_0=None,
        t_E=None,
        times=None,
        filters=None,
        rng_info=None,
        **kwargs,
    ):
        """Apply the effect to band fluxes.

        Parameters
        ----------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD).
        filters : numpy.ndarray, optional
            A length N array of filters. If not provided, the effect is applied to all
            band fluxes.
        apply_microlensing : bool
            Whether or not to apply microlensing.
        microlensing_t0 : float
            The time of the microlensing peak, in days.
        u_0 : float
            Impact parameter, distance from the source at time
            of peak in Einstein radii.
        t_E : float
            Einstein timescale, in days.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes after the effect is applied (in nJy).
        """
        # If we do not apply microlensing, we can skip the entire computation.
        if not apply_microlensing:
            return bandfluxes

        if microlensing_t0 is None:
            raise ValueError("microlensing_t0 must be provided")
        if u_0 is None:
            raise ValueError("u_0 must be provided")
        if t_E is None:
            raise ValueError("t_E must be provided")

        # Array of parameters
        # Note that VBMicrolensing requires some parameters
        # in log space, as per example:
        # https://github.com/valboz/VBMicrolensing/blob/main/docs/python/LightCurves.md
        pr = [np.log(u_0), np.log(t_E), microlensing_t0]

        # Calculates the PSPL magnification at different times with parameters in pr
        # Attention: VBMicrolensing assumes that times is an np.array or a list.
        vbm_results = self.VBM.PSPLLightCurve(pr, times)

        # array of magnifications at each time in time_stamp
        bandfluxes = bandfluxes * np.asarray(vbm_results[0])

        return bandfluxes



from pathlib import Path

# Define some global directory paths to use for testing, notebooks, etc.
_LIGHTCURVELYNX_BASE_DIR = Path(__file__).parent.parent.parent
_LIGHTCURVELYNX_BASE_DATA_DIR = _LIGHTCURVELYNX_BASE_DIR / "data"

_LIGHTCURVELYNX_TEST_DIR = _LIGHTCURVELYNX_BASE_DIR / "tests" / "lightcurvelynx"
_LIGHTCURVELYNX_TEST_DATA_DIR = _LIGHTCURVELYNX_TEST_DIR / "data"



"""The core functions for running the LightCurveLynx simulation."""

import citation_compass as cc
import numpy as np
import pandas as pd
from nested_pandas import NestedFrame

from lightcurvelynx.astro_utils.noise_model import apply_noise
from lightcurvelynx.models.physical_model import BandfluxModel


def get_time_windows(t0, time_window_offset):
    """Get the time windows for each sample state based on the time window offset.

    Parameters
    ----------
    t0 : float or np.ndarray, optional
        The reference time (t0) for the time windows.
    time_window_offset : tuple(float, float), optional
        A tuple specifying the time window offset (before, after) t0 in days.
        If None, no time window is applied.

    Returns
    -------
    start_times : np.ndarray or None
        The start times for each sample t0 - time_window_offset[0]. If a before time is given,
        this is always returned as an array (even if t0 is a scalar). None returned if there is
        no start time.
    end_times : np.ndarray or None
        The end times for each sample t0 - time_window_offset[1]. If an after time is given,
        this is always returned as an array (even if t0 is a scalar). None returned if there is
        no end time.
    """
    # If the model did not have a t0 or we do not have a time_window_offset,
    # we cannot apply a time window.
    if t0 is None or time_window_offset is None:
        return None, None
    if len(time_window_offset) != 2:
        raise ValueError("time_window_offset must be a tuple of (before, after) in days.")
    before, after = time_window_offset

    # If t0 is a scalar apply the offset directly.
    if np.isscalar(t0):
        t0 = np.array([t0])
    start_times = t0 - before if before is not None else None
    end_times = t0 + after if after is not None else None

    return start_times, end_times


def simulate_lightcurves(
    model,
    num_samples,
    obstable,
    passbands,
    *,
    obstable_save_cols=None,
    param_cols=None,
    apply_obs_mask=False,
    time_window_offset=None,
    rng=None,
    generate_citations=False,
):
    """Generate a number of simulations of the given model and information
    from one or more surveys.

    Parameters
    ----------
    model : BasePhysicalModel
        The model to draw from. This may have its own parameters which
        will be randomly sampled with each draw.
    num_samples : int
        The number of samples.
    obstable : ObsTable or List of ObsTable
        The ObsTable(s) from which to extract information for the samples.
    passbands : PassbandGroup or List of PassbandGroup
        The passbands to use for generating the bandfluxes.
    apply_obs_mask: boolean
        If True, apply obs_mask to filter interesting indices/times.
    time_window_offset : tuple(float, float), optional
        A tuple specifying the time window offset (before, after) t0 in days.
        This is used to filter the observations to only those within the specified
        time window (t0 - before, t0 + after). If None or the model does not have a
        t0 specified, no time window is applied.
    obstable_save_cols : list of str, optional
        A list of ObsTable columns to be saved as part of the results. This is used
        to save context information about how the light curves were generated. If the column
        is missing from one of the ObsTables, a null value such as None or NaN is used.
        If None, no additional columns are saved.
    param_cols : list of str, optional
        A list of the model's parameter columns to be saved as separate columns in
        the results (instead of just the full dictionary of parameters). These
        must be specified as strings in the node_name.param_name format.
        If None, no additional columns are saved.
    rng : numpy.random._generator.Generator, optional
        A given numpy random number generator to use for this computation. If not
        provided, the function uses the node's random number generator.
    generate_citations : bool, optional
        If True, generate citations for the simulation and output them.

    Returns
    -------
    lightcurves : nested_pandas.NestedFrame
        A NestedFrame with a row for each object.
    """
    # Sample the parameter space of this model. We do this once for all surveys, so the
    # object use the same parameters across all observations.
    if num_samples <= 0:
        raise ValueError("Invalid number of samples.")
    sample_states = model.sample_parameters(num_samples=num_samples, rng_info=rng)

    # If we are given information for a single survey, make it into a list.
    if not isinstance(obstable, list):
        obstable = [obstable]
    if not isinstance(passbands, list):
        passbands = [passbands]
    num_surveys = len(obstable)
    if num_surveys != len(passbands):
        raise ValueError("Number of surveys must match number of passbands.")

    # We do not currently support bandflux models with multiple surveys because
    # a bandflux model is defined relative to a single survey.
    if num_surveys > 1 and isinstance(model, BandfluxModel):
        raise ValueError(
            "Simulating a BandfluxModel with multiple surveys is currently not supported, "
            "because the bandflux model is defined relative to the filters of a single survey."
        )

    # Create a dictionary for the object level information, including any saved parameters.
    # Some of these are placeholders (e.g. nobs) until they can be filled in during the simulation.
    ra = np.atleast_1d(model.get_param(sample_states, "ra"))
    dec = np.atleast_1d(model.get_param(sample_states, "dec"))
    results_dict = {
        "id": [i for i in range(num_samples)],
        "ra": ra.tolist(),
        "dec": dec.tolist(),
        "nobs": [0] * num_samples,
        "t0": np.atleast_1d(model.get_param(sample_states, "t0")).tolist(),
        "z": np.atleast_1d(model.get_param(sample_states, "redshift")).tolist(),
        "params": [state.to_dict() for state in sample_states],
    }
    if param_cols is not None:
        for col in param_cols:
            if col not in sample_states:
                raise KeyError(
                    f"Parameter column {col} not found in model parameters. "
                    f"Available parameters are: {sample_states.get_all_params_names()}."
                )
            results_dict[col.replace(".", "_")] = np.atleast_1d(sample_states[col]).tolist()

    # Set up the nested array for the per-observation data, including ObsTable information.
    nested_index = []
    nested_dict = {"mjd": [], "filter": [], "flux": [], "fluxerr": [], "flux_perfect": [], "survey_idx": []}
    if obstable_save_cols is None:
        obstable_save_cols = []
    for col in obstable_save_cols:
        nested_dict[col] = []

    # Determine which of the of the simulated positions match the pointings from each ObsTable.
    start_times, end_times = get_time_windows(
        model.get_param(sample_states, "t0"),
        time_window_offset,
    )
    all_obs_matches = [
        obstable[i].range_search(ra, dec, t_min=start_times, t_max=end_times) for i in range(num_surveys)
    ]

    # Get all times and all filters as numpy arrays so we can do easy subsets.
    all_times = [np.asarray(obstable[i]["time"].values, dtype=float) for i in range(num_surveys)]
    all_filters = [np.asarray(obstable[i]["filter"].values, dtype=str) for i in range(num_surveys)]

    # We loop over objects first, then surveys. This allows us to generate a single block
    # of data for the object over all surveys.
    for idx, state in enumerate(sample_states):
        total_num_obs = 0

        for survey_idx in range(num_surveys):
            # Find the indices and times where the current model is seen.
            obs_index = np.asarray(all_obs_matches[survey_idx][idx])
            if len(obs_index) == 0:
                obs_times = []
                obs_filters = []
            else:
                obs_times = all_times[survey_idx][obs_index]

                # Filter to only the "interesting" indices / times for this object.
                if apply_obs_mask:
                    obs_mask = model.mask_by_time(obs_times, state)
                    obs_index = obs_index[obs_mask]
                    obs_times = obs_times[obs_mask]
                # Extract the filters for this observation.
                obs_filters = all_filters[survey_idx][obs_index]

            # Compute the bandfluxes and errors over just the given filters.
            bandfluxes_perfect = model.evaluate_bandfluxes(
                passbands[survey_idx], obs_times, obs_filters, state
            )
            bandfluxes_error = obstable[survey_idx].bandflux_error_point_source(bandfluxes_perfect, obs_index)
            bandfluxes = apply_noise(bandfluxes_perfect, bandfluxes_error, rng=rng)

            # Append the per-observation data to the nested dictionary, including
            # and needed ObsTable columns.
            nobs = len(obs_times)
            nested_dict["mjd"].extend(list(obs_times))
            nested_dict["filter"].extend(list(obs_filters))
            nested_dict["flux_perfect"].extend(list(bandfluxes_perfect))
            nested_dict["flux"].extend(list(bandfluxes))
            nested_dict["fluxerr"].extend(list(bandfluxes_error))
            nested_dict["survey_idx"].extend([survey_idx] * nobs)
            for col in obstable_save_cols:
                col_data = (
                    list(obstable[survey_idx][col].values[obs_index])
                    if col in obstable[survey_idx]
                    else [None] * nobs
                )
                nested_dict[col].extend(col_data)

            total_num_obs += nobs
            nested_index.extend([idx] * nobs)

        # The number of observations is the total across all surveys.
        results_dict["nobs"][idx] = total_num_obs

    # Create the nested frame.
    results = NestedFrame(data=results_dict, index=[i for i in range(num_samples)])
    nested_frame = pd.DataFrame(data=nested_dict, index=nested_index)
    results = results.add_nested(nested_frame, "lightcurve")

    # If requested, generate citations for the simulation.
    if generate_citations:
        print(
            "The following citations were called during this simulation. Note that this list is "
            "provided for convenience and may be incomplete and we recommend the user confirm "
            "which models, effects, and parameters were used."
        )
        cc.print_used_citations()

    return results


def compute_single_noise_free_lightcurve(
    model,
    graph_state,
    passbands,
    *,
    rest_frame_phase_min=-50.0,
    rest_frame_phase_max=50.0,
    rest_frame_phase_step=2.0,
):
    """Compute the noise-free light curve for a single object.

    This is a helper function for compute_noise_free_lightcurves.

    Parameters
    ----------
    model : BasePhysicalModel
        The model object to use for generating the light curves.
    graph_state : GraphState
        The state of the graph for the simulation. Must be a single state
        (num_samples=1).
    passbands : PassbandGroup
        The passbands to use for generating the bandfluxes.
    rest_frame_phase_min : float or np.ndarray
        The minimum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is -50.0 days.
    rest_frame_phase_max : float or np.ndarray
        The maximum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 50.0 days.
    rest_frame_phase_step : float or np.ndarray
        The step size (in days) between rest-frame phases at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 2.0 days.

    Returns
    -------
    lightcurves : dict
        A dictionary mapping each filter name to the corresponding array of bandfluxes (in nJy),
        with an additional keys "times" (in MJD) and "rest_phase" (in days relative to t0).
    """
    if graph_state.num_samples != 1:
        raise ValueError("graph_state must have num_samples=1.")

    # Generate the rest-frame times at which to evaluate the light curve.
    rest_phase = np.arange(rest_frame_phase_min, rest_frame_phase_max, step=rest_frame_phase_step)
    times = rest_phase.copy()

    # Compute the observed-frame times (MJD) if needed.
    redshift = model.get_param(graph_state, "redshift")
    if redshift is not None and redshift > 0:
        times = times * (1 + redshift)

    t0 = model.get_param(graph_state, "t0")
    if t0 is not None:
        times += t0

    # Compute the light curve without noise for each time in each filter.
    lightcurves = {"times": times, "rest_phase": rest_phase}
    for filter in passbands.filters:
        filters_array = np.full(len(times), filter)
        bandfluxes = model.evaluate_bandfluxes(passbands, times, filters_array, graph_state)
        lightcurves[filter] = bandfluxes
    return lightcurves


def compute_noise_free_lightcurves(
    model,
    graph_state,
    passbands,
    *,
    rest_frame_phase_min=-50.0,
    rest_frame_phase_max=50.0,
    rest_frame_phase_step=2.0,
):
    """Compute the noise-free light curves for a given model and one more more states
    at given times (in either MJD or rest-frame phase).

    This function simulates the light curves without adding any noise, allowing
    for the analysis of the underlying model behavior.

    Parameters
    ----------
    model : BasePhysicalModel
        The model object to use for generating the light curves.
    graph_state : GraphState
        The state of the graph for the simulation.
    passbands : PassbandGroup
        The passbands to use for generating the bandfluxes.
    rest_frame_phase_min : float or np.ndarray
        The minimum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is -50.0 days.
    rest_frame_phase_max : float or np.ndarray
        The maximum rest-frame phase (in days) at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 50.0 days.
    rest_frame_phase_step : float or np.ndarray
        The step size (in days) between rest-frame phases at which to evaluate the light curve.
        If an array is given, it must match the number of samples in graph_state.
        Default is 2.0 days.

    Returns
    -------
    lightcurves : nested_pandas.NestedFrame
        A NestedFrame with a row for each object that contains a "lightcurve" column
        with a dictionary mapping each filter name to the corresponding array of
        bandfluxes (in nJy), with an additional nested columns "times" (in MJD) and
        "rest_phase" (in days relative to t0).
    """
    # Expand any scalar inputs to arrays.
    num_samples = graph_state.num_samples
    if np.isscalar(rest_frame_phase_min):
        rest_frame_phase_min = np.full(num_samples, rest_frame_phase_min)
    if np.isscalar(rest_frame_phase_max):
        rest_frame_phase_max = np.full(num_samples, rest_frame_phase_max)
    if np.isscalar(rest_frame_phase_step):
        rest_frame_phase_step = np.full(num_samples, rest_frame_phase_step)

    # Set up the per-model dictionary.
    results_dict = {
        "id": np.arange(num_samples).tolist(),
        "ra": np.atleast_1d(model.get_param(graph_state, "ra")).tolist(),
        "dec": np.atleast_1d(model.get_param(graph_state, "dec")).tolist(),
        "z": np.atleast_1d(model.get_param(graph_state, "redshift")).tolist(),
    }

    # Set up the nested dictionary with the light curve information.
    all_filters = passbands.filters
    nested_index = []
    nested_dict = {"times": [], "rest_phase": []}
    for filter in all_filters:
        nested_dict[filter] = []

    # Compute each light curve.
    for idx, state_i in enumerate(graph_state):
        lc = compute_single_noise_free_lightcurve(
            model,
            state_i,
            passbands,
            rest_frame_phase_min=rest_frame_phase_min[idx],
            rest_frame_phase_max=rest_frame_phase_max[idx],
            rest_frame_phase_step=rest_frame_phase_step[idx],
        )

        # Append the light curve data onto the nested dictionary.
        for key in nested_dict:
            nested_dict[key].extend(list(lc[key]))
        nested_index.extend([idx] * len(lc["times"]))

    # Create the nested results frame.
    results = NestedFrame(data=results_dict, index=[i for i in range(num_samples)])
    nested_frame = pd.DataFrame(data=nested_dict, index=nested_index)
    results = results.add_nested(nested_frame, "lightcurve")
    return results



import numpy as np
import pytest
from lightcurvelynx.effects.effect_model import EffectModel


def test_effect_model() -> None:
    """The base effect model is able to extract the parameters."""
    model = EffectModel(param1=1.0, param2=2.0)
    assert "param1" in model.parameters
    assert "param2" in model.parameters
    assert model.parameters["param1"] == 1.0
    assert model.parameters["param2"] == 2.0
    assert model.effect_name == "EffectModel"
    assert str(model) == "EffectModel"
    assert repr(model) == "EffectModel(param1,param2)"

    # We can override the effect's name.
    model2 = EffectModel(param3=3.0, effect_name="test_effect")
    assert str(model2) == "test_effect"
    assert repr(model2) == "test_effect(param3)"

    # We cannot call apply on the base EffectModel
    with pytest.raises(NotImplementedError):
        _ = model.apply(np.zeros((5, 3)))



import tempfile
from pathlib import Path

import astropy.units as u
import numpy as np
import pandas as pd
import pytest
from astropy.coordinates import SkyCoord
from lightcurvelynx.obstable.obs_table import ObsTable
from regions import CircleSkyRegion


def test_create_obs_table():
    """Create a minimal ObsTable object and perform basic queries."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
    }
    pdf = pd.DataFrame(values)

    ops_data = ObsTable(pdf)
    assert len(ops_data) == 5
    assert len(ops_data.columns) == 4

    # We have all the attributes set at their default values (which are None for the base class).
    assert ops_data.survey_values["dark_current"] is None
    assert ops_data.survey_values["ext_coeff"] is None
    assert ops_data.survey_values["pixel_scale"] is None
    assert ops_data.survey_values["radius"] is None
    assert ops_data.survey_values["read_noise"] is None
    assert ops_data.survey_values["zp_per_sec"] is None
    assert ops_data.survey_values["survey_name"] == "Unknown"

    # Check that we can extract the time bounds.
    t_min, t_max = ops_data.time_bounds()
    assert t_min == 0.0
    assert t_max == 4.0

    # We can query which columns the ObsTable has.
    assert "time" in ops_data.columns
    assert "ra" in ops_data.columns
    assert "dec" in ops_data.columns
    assert "is_good_obs" not in ops_data.columns

    # We can access columns directly as though it was a table.
    assert np.allclose(ops_data["ra"], values["ra"])
    assert np.allclose(ops_data["dec"], values["dec"])
    assert np.allclose(ops_data["time"], values["time"])

    # Without a filters column we cannot access the filters.
    assert len(ops_data.filters) == 0

    # We can create an ObsTable directly from the dictionary as well.
    ops_data2 = ObsTable(pdf)
    assert len(ops_data2) == 5
    assert len(ops_data.columns) == 4
    assert np.allclose(ops_data2["ra"], values["ra"])
    assert np.allclose(ops_data2["dec"], values["dec"])
    assert np.allclose(ops_data2["time"], values["time"])

    # We raise an error if we are missing a required row.
    del values["dec"]
    with pytest.raises(KeyError):
        _ = ObsTable(values)


def test_create_obs_table_override():
    """Test that we can override the default obs table values."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    ops_data = ObsTable(
        values,
        dark_current=0.1,
        ext_coeff={"u": 0.1, "g": 0.2, "r": 0.3, "i": 0.4, "z": 0.5, "y": 0.6},
        pixel_scale=0.1,
        radius=1.0,
        read_noise=5.0,
        zp_per_sec={"u": 25.0, "g": 26.0, "r": 27.0, "i": 28.0, "z": 29.0, "y": 30.0},
    )

    # We have loaded the non-default values.
    assert ops_data.survey_values["dark_current"] == 0.1
    assert ops_data.survey_values["ext_coeff"] == {"u": 0.1, "g": 0.2, "r": 0.3, "i": 0.4, "z": 0.5, "y": 0.6}
    assert ops_data.survey_values["pixel_scale"] == 0.1
    assert ops_data.survey_values["radius"] == 1.0
    assert ops_data.survey_values["read_noise"] == 5.0
    assert ops_data.survey_values["zp_per_sec"] == {
        "u": 25.0,
        "g": 26.0,
        "r": 27.0,
        "i": 28.0,
        "z": 29.0,
        "y": 30.0,
    }

    # We can access the filters.
    assert set(ops_data.filters) == {"r", "g", "i"}

    # Check that we can read in the defaults from the pandas metadata.
    pdf = pd.DataFrame(values)
    pdf.attrs["lightcurvelynx_survey_data"] = {
        "survey_name": "test",
        "pixel_scale": 0.1,
        "dark_current": 0.2,
        "radius": 2.0,
    }
    ops_data2 = ObsTable(pdf, dark_current=0.5, radius=1.0)

    # Check that we use use the updated defaults, preferring the keyword arguments
    # to the table's metadata.
    assert ops_data2.survey_values["survey_name"] == "test"
    assert ops_data2.survey_values["pixel_scale"] == 0.1
    assert ops_data2.survey_values["dark_current"] == 0.5
    assert ops_data2.survey_values["radius"] == 1.0

    # We can use the radius to get and set the property.
    assert ops_data2.radius == 1.0
    ops_data2.radius = 5.0
    assert ops_data2.survey_values["radius"] == 5.0

    # If we delete the radius survey value, we get None and can reset it.
    del ops_data2.survey_values["radius"]
    assert ops_data2.radius is None
    ops_data2.radius = 5.0
    assert ops_data2.survey_values["radius"] == 5.0


def test_create_obs_table_custom_names():
    """Create a minimal ObsTable object from alternate column names."""
    values = {
        "custom_time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "custom_ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "custom_dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
        "other": np.zeros(5),
    }

    # Load fails if we use the default colmap.
    with pytest.raises(KeyError):
        _ = ObsTable(values)

    # Load succeeds if we pass in a customer dictionary.
    colmap = {"ra": "custom_ra", "dec": "custom_dec", "time": "custom_time"}
    ops_data = ObsTable(values, colmap=colmap)
    assert len(ops_data) == 5

    # Test that we fail if we try to map a column onto an existing column (mapping the given
    # "other" column onto the existing "zp" column).
    colmap = {"ra": "custom_ra", "dec": "custom_dec", "time": "custom_time", "zp": "other"}
    with pytest.raises(ValueError):
        _ = ObsTable(values, colmap=colmap)


def test_obs_table_add_columns():
    """Create a minimal ObsTable object and perform basic queries."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
    }
    pdf = pd.DataFrame(values)

    ops_data = ObsTable(pdf)
    assert len(ops_data) == 5
    assert len(ops_data.columns) == 4

    # We can add a column of constant values from a scalar.
    assert "new_column1" not in ops_data.columns
    ops_data.add_column("new_column1", 10)
    assert "new_column1" in ops_data.columns
    assert np.allclose(ops_data["new_column1"], [10, 10, 10, 10, 10])

    # We can add a column of data from a list or array.
    assert "new_column2" not in ops_data.columns
    ops_data.add_column("new_column2", [1, 2, 3, 4, 5])
    assert "new_column2" in ops_data.columns
    assert np.allclose(ops_data["new_column2"], [1, 2, 3, 4, 5])

    # We fail if we add a column with the incorrect number of values.
    with pytest.raises(ValueError):
        ops_data.add_column("new_column3", [1, 2, 3, 4, 5, 6])

    # We fail if we try to overwrite a current column unless we
    # set overwrite=True.
    with pytest.raises(KeyError):
        ops_data.add_column("new_column1", 12)
    assert np.allclose(ops_data["new_column1"], [10, 10, 10, 10, 10])

    ops_data.add_column("new_column2", 12, overwrite=True)
    assert np.allclose(ops_data["new_column2"], [12, 12, 12, 12, 12])


def test_obs_table_filter_rows():
    """Test that the user can filter out ObsTable rows."""
    times = np.arange(0.0, 10.0, 1.0)
    values = {
        "time": times,
        "ra": 15.0 * (times + 1.0),
        "dec": -1.0 * times,
        "zp": np.ones(10),
        "filter": np.tile(["r", "g"], 5),
        "custom_column": np.zeros(10),
    }
    ops_data = ObsTable(values, dark_current=0.1, colmap={"custom_column": "my_col"})
    assert len(ops_data) == 10
    assert len(ops_data.columns) == 6
    assert ops_data.survey_values["dark_current"] == 0.1

    # We can filter the ObsTable to specific rows by index.
    inds = [0, 1, 2, 3, 4, 5, 7, 8]
    ops_data = ops_data.filter_rows(inds)
    assert len(ops_data) == 8
    assert len(ops_data.columns) == 6

    # Check that we propagate the survey values and the column mapping.
    assert ops_data.survey_values["dark_current"] == 0.1
    assert "custom_column" in ops_data
    assert "my_col" in ops_data

    expected_times = np.array(inds)
    assert np.allclose(ops_data["time"], expected_times)
    assert np.allclose(ops_data["ra"], 15.0 * (expected_times + 1.0))
    assert np.allclose(ops_data["dec"], -1.0 * expected_times)
    assert np.array_equal(ops_data["filter"], values["filter"][inds])

    # Check that the size of the internal KD-tree has changed.
    assert ops_data._kd_tree.n == 8

    # We can filter the OpSim to specific rows by mask.
    ops_data = ops_data.filter_rows(ops_data["filter"] == "r")
    assert len(ops_data) == 4

    expected_times = np.array([0.0, 2.0, 4.0, 8.0])
    assert np.allclose(ops_data["time"], expected_times)
    assert np.allclose(ops_data["ra"], 15.0 * (expected_times + 1.0))
    assert np.allclose(ops_data["dec"], -1.0 * expected_times)
    assert np.all(ops_data["filter"] == "r")

    # Check that the size of the internal KD-tree has changed (again).
    assert ops_data._kd_tree.n == 4

    # We throw an error if the mask is the wrong size.
    bad_mask = [True] * (len(ops_data) - 1)
    with pytest.raises(ValueError):
        _ = ops_data.filter_rows(bad_mask)


def test_read_small_obs_table(opsim_small):
    """Read in a small ObsTable file from the testing data directory."""
    colmap = {
        "airmass": "airmass",
        "dec": "fieldDec",
        "exptime": "visitExposureTime",
        "filter": "filter",
        "ra": "fieldRA",
        "time": "observationStartMJD",
        "zp": "zp_nJy",  # We add this column to the table
        "seeing": "seeingFwhmEff",
        "skybrightness": "skyBrightness",
        "nexposure": "numExposures",
    }
    ops_data = ObsTable.from_db(opsim_small, colmap=colmap)
    assert len(ops_data) == 300


def test_write_read_obs_table_parquet():
    """Create a minimal observation table data frame, test that we can write it as
    a parquet file, and test that we can correctly read it back in."""

    # Create a fake obstable data frame with just time, RA, dec, and zp. Add two
    # pieces of metadata (survey name and pixel scale).
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
    }
    ops_data = ObsTable(pd.DataFrame(values), survey_name="test", pixel_scale=0.1)

    with tempfile.TemporaryDirectory() as dir_name:
        file_path = Path(dir_name, "test_write_read_obs_table.parquet")

        # The obstable does not exist until we write it.
        assert not file_path.is_file()
        with pytest.raises(FileNotFoundError):
            _ = ObsTable.from_parquet(file_path)

        # We can write the obstable parquet.
        ops_data.write_parquet(file_path)
        assert file_path.is_file()

        # We can reread the obstable parquet.
        ops_data2 = ObsTable.from_parquet(file_path)
        assert len(ops_data2) == 5
        assert np.allclose(values["time"], ops_data2["time"].to_numpy())
        assert np.allclose(values["ra"], ops_data2["ra"].to_numpy())
        assert np.allclose(values["dec"], ops_data2["dec"].to_numpy())
        assert ops_data2.survey_values["survey_name"] == "test"
        assert ops_data2.survey_values["pixel_scale"] == 0.1

        # We cannot overwrite unless we set overwrite=True
        with pytest.raises(FileExistsError):
            ops_data.write_parquet(file_path, overwrite=False)
        ops_data.write_parquet(file_path, overwrite=True)


def test_write_read_obs_table():
    """Create a minimal observation table data frame, test that we can write it,
    and test that we can correctly read it back in."""

    # Create a fake obstable data frame with just time, RA, and dec.
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
    }
    ops_data = ObsTable(pd.DataFrame(values))

    with tempfile.TemporaryDirectory() as dir_name:
        file_path = Path(dir_name, "test_write_read_obs_table.db")

        # The obstable does not exist until we write it.
        assert not file_path.is_file()
        with pytest.raises(FileNotFoundError):
            _ = ObsTable.from_db(file_path)

        # We can write the obstable db.
        ops_data.write_db(file_path)
        assert file_path.is_file()

        # We can reread the obstable db.
        ops_data2 = ObsTable.from_db(file_path)
        assert len(ops_data2) == 5
        assert np.allclose(values["time"], ops_data2["time"].to_numpy())
        assert np.allclose(values["ra"], ops_data2["ra"].to_numpy())
        assert np.allclose(values["dec"], ops_data2["dec"].to_numpy())

        # We cannot overwrite unless we set overwrite=True
        with pytest.raises(ValueError):
            ops_data.write_db(file_path, overwrite=False)
        ops_data.write_db(file_path, overwrite=True)


def test_obs_table_range_search():
    """Test that we can extract the time, ra, and dec from an obs table data frame."""
    # Create a fake obs table data frame with just time, RA, and dec.
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]),
        "ra": np.array([15.0, 15.0, 15.01, 15.0, 25.0, 24.99, 60.0, 5.0]),
        "dec": np.array([-10.0, 10.0, 10.01, 9.99, 10.0, 9.99, -5.0, -1.0]),
        "zp": np.ones(8),
    }
    ops_data = ObsTable(values)

    # Test single queries.
    assert set(ops_data.range_search(15.0, 10.0, radius=0.5)) == set([1, 2, 3])
    assert set(ops_data.range_search(25.0, 10.0, radius=0.5)) == set([4, 5])
    assert set(ops_data.range_search(15.0, 10.0, radius=100.0)) == set([0, 1, 2, 3, 4, 5, 6, 7])
    assert set(ops_data.range_search(15.0, 10.0, radius=1e-6)) == set([1])
    assert set(ops_data.range_search(15.02, 10.0, radius=1e-6)) == set()

    # Test that we can filter by time.
    assert set(ops_data.range_search(15.0, 10.0, radius=0.5, t_min=1.0, t_max=3.0)) == set([1, 2, 3])
    assert set(ops_data.range_search(15.0, 10.0, radius=0.5, t_min=2.0, t_max=4.0)) == set([2, 3])
    assert set(ops_data.range_search(15.0, 10.0, radius=0.5, t_min=0.0, t_max=1.0)) == set([1])
    assert set(ops_data.range_search(15.0, 10.0, radius=0.5, t_min=4.0, t_max=5.0)) == set()

    # With no radius provided (and no default), the query fails.
    with pytest.raises(ValueError):
        _ = ops_data.range_search(15.0, 10.0)

    # Test is_observed() with single queries.
    assert ops_data.is_observed(15.0, 10.0, radius=0.5)
    assert not ops_data.is_observed(15.02, 10.0, radius=1e-6)
    assert ops_data.is_observed(15.0, 10.0, radius=0.5, t_min=1.0, t_max=3.0)
    assert not ops_data.is_observed(15.0, 10.0, radius=0.5, t_min=40.0, t_max=50.0)

    # Test a batched query.
    query_ra = np.array([15.0, 25.0, 15.0])
    query_dec = np.array([10.0, 10.0, 5.0])
    neighbors = ops_data.range_search(query_ra, query_dec, radius=0.5)
    assert len(neighbors) == 3
    assert set(neighbors[0]) == set([1, 2, 3])
    assert set(neighbors[1]) == set([4, 5])
    assert set(neighbors[2]) == set()

    # Do the same query with time filtering.
    t_min = np.array([0.0, 5.0, 0.0])
    t_max = np.array([2.0, 11.0, 1.0])
    neighbors = ops_data.range_search(query_ra, query_dec, radius=0.5, t_min=t_min, t_max=t_max)
    assert len(neighbors) == 3
    assert set(neighbors[0]) == set([1, 2])
    assert set(neighbors[1]) == set([5])
    assert set(neighbors[2]) == set()

    # Test is_observed() with batched queries.
    assert np.array_equal(
        ops_data.is_observed(query_ra, query_dec, radius=0.5),
        np.array([True, True, False]),
    )

    # Test that we fail if bad query arrays are provided.
    with pytest.raises(ValueError):
        _ = ops_data.range_search(None, None, radius=0.5)
    with pytest.raises(ValueError):
        _ = ops_data.range_search([1.0, 2.3], 4.5, radius=0.5)
    with pytest.raises(ValueError):
        _ = ops_data.range_search([1.0, 2.3], [4.5, 6.7, 8.9], radius=0.5)
    with pytest.raises(ValueError):
        _ = ops_data.range_search([1.0, 2.3], [4.5, None], radius=0.5)


def test_obs_table_get_observations():
    """Test that we can extract the time, ra, and dec from an obs table data frame."""
    # Create a fake obs table data frame with just time, RA, and dec.
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]),
        "ra": np.array([15.0, 15.0, 15.01, 15.0, 25.0, 24.99, 60.0, 5.0]),
        "dec": np.array([-10.0, 10.0, 10.01, 9.99, 10.0, 9.99, -5.0, -1.0]),
        "zp": np.ones(8),
        "airmass_data": np.array([1, 2, 3, 4, 5, 6, 7, 8]),
    }
    ops_data = ObsTable(values, colmap={"airmass_data": "airmass"})

    # Test basic queries (all columns).
    obs = ops_data.get_observations(15.0, 10.0, radius=0.5)
    assert len(obs) == 5
    assert np.allclose(obs["time"], [1.0, 2.0, 3.0])

    obs = ops_data.get_observations(25.0, 10.0, radius=0.5)
    assert np.allclose(obs["time"], [4.0, 5.0])

    obs = ops_data.get_observations(15.0, 10.0, radius=100.0)
    assert np.allclose(obs["time"], [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0])

    obs = ops_data.get_observations(15.0, 10.0, radius=1e-6)
    assert np.allclose(obs["time"], [1.0])

    obs = ops_data.get_observations(15.02, 10.0, radius=1e-6)
    assert len(obs["time"]) == 0

    # Test we can get a subset of columns.
    obs = ops_data.get_observations(15.0, 10.0, radius=0.5, cols=["time", "zp"])
    assert len(obs) == 2
    assert np.allclose(obs["time"], [1.0, 2.0, 3.0])

    # Test we can use the mapped or original column names.
    obs = ops_data.get_observations(15.0, 10.0, radius=0.5, cols="airmass")
    assert np.allclose(obs["airmass"], [2, 3, 4])
    obs = ops_data.get_observations(15.0, 10.0, radius=0.5, cols="airmass_data")
    assert np.allclose(obs["airmass_data"], [2, 3, 4])

    # Test we fail with an unrecognized column name.
    with pytest.raises(KeyError):
        _ = ops_data.get_observations(15.0, 10.0, radius=0.5, cols=["time", "custom_col"])


def test_obs_table_range_search_detector_footprint():
    """Test that we can extract the time, ra, and dec from an obs table data frame
    with a detector footprint."""
    # Create a fake obs table data frame with just time, RA, and dec.
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]),
        "ra": np.array([15.0, 15.0, 15.01, 15.0, 25.0, 24.99, 60.0, 5.0]),
        "dec": np.array([-10.0, 10.0, 10.01, 9.99, 10.0, 9.99, -5.0, -1.0]),
        "zp": np.ones(8),
    }
    # Add a circular footprint with radius 0.5 deg.
    detector_footprint = CircleSkyRegion(center=SkyCoord(ra=0.0, dec=0.0, unit="deg"), radius=0.5 * u.deg)
    ops_data = ObsTable(values, detector_footprint=detector_footprint, pixel_scale=0.1)

    # Test single queries. Despite the given radius, we only find points within the detector footprint.
    assert set(ops_data.range_search(15.0, 10.0, radius=2.0)) == set([1, 2, 3])
    assert set(ops_data.range_search(25.0, 10.0, radius=2.0)) == set([4, 5])
    assert set(ops_data.range_search(15.0, 10.0, radius=100.0)) == set([1, 2, 3])
    assert set(ops_data.range_search(15.0, 10.0, radius=1e-6)) == set([1])
    assert set(ops_data.range_search(15.02, 10.0, radius=1e-6)) == set()

    # Test a batched queries and that the detector footprint is applied to all of them.
    query_ra = np.array([15.0, 25.0, 15.0])
    query_dec = np.array([10.0, 10.0, 5.0])
    neighbors = ops_data.range_search(query_ra, query_dec, radius=10.0)
    assert len(neighbors) == 3
    assert set(neighbors[0]) == set([1, 2, 3])
    assert set(neighbors[1]) == set([4, 5])
    assert set(neighbors[2]) == set()

    # We can shut off the detector footprint to save time and use the approximate radius.
    ops_data.clear_detector_footprint()
    assert set(ops_data.range_search(15.0, 10.0, radius=100.0)) == set([0, 1, 2, 3, 4, 5, 6, 7])


def test_obs_table_docstring():
    """Test if ObsTable class has a docstring"""
    assert ObsTable.__doc__ is not None
    assert len(ObsTable.__doc__) > 100


def test_read_obs_table_shorten(opsim_shorten):
    """Read in a shorten ObsTable file from the testing data directory."""
    colmap = {
        "airmass": "airmass",
        "dec": "fieldDec",
        "exptime": "visitExposureTime",
        "filter": "filter",
        "ra": "fieldRA",
        "time": "observationStartMJD",
        "zp": "zp_nJy",  # We add this column to the table
        "seeing": "seeingFwhmEff",
        "skybrightness": "skyBrightness",
        "nexposure": "numExposures",
    }
    ops_data = ObsTable.from_db(opsim_shorten, colmap=colmap)
    assert len(ops_data) == 100


def test_build_moc():
    """Test building a Multi-Order Coverage Map (MOC) from the ObsTable."""
    # Create an ObsTable with three pointings.
    values = {
        "time": np.array([0.0, 1.0, 2.0]),
        "ra": np.array([10.0, 11.0, 45.0]),
        "dec": np.array([0.0, 0.0, -10.0]),
        "zp": np.ones(3),
    }
    ops_data = ObsTable(values)

    # We fail if no radius is given (including no default.)
    with pytest.raises(ValueError):
        _ = ops_data.build_moc()

    moc = ops_data.build_moc(radius=1.5)
    assert moc is not None
    assert moc.max_order == 10

    # Test that the MOC covers the correct area.
    assert moc.contains_skycoords(SkyCoord(ra=10.0, dec=0.0, unit="deg"))
    assert moc.contains_skycoords(SkyCoord(ra=11.0, dec=1.0, unit="deg"))
    assert moc.contains_skycoords(SkyCoord(ra=10.5, dec=0.5, unit="deg"))
    assert moc.contains_skycoords(SkyCoord(ra=45.0, dec=-10.0, unit="deg"))
    assert moc.contains_skycoords(SkyCoord(ra=45.01, dec=-9.99, unit="deg"))
    assert moc.contains_skycoords(SkyCoord(ra=44.98, dec=-10.01, unit="deg"))
    assert not moc.contains_skycoords(SkyCoord(ra=46.00, dec=0.0, unit="deg"))
    assert not moc.contains_skycoords(SkyCoord(ra=14.00, dec=0.0, unit="deg"))
    assert not moc.contains_skycoords(SkyCoord(ra=52.00, dec=-10.0, unit="deg"))
